<?php

namespace Oro\Bundle\DataGridBundle\Tests\Behat\Element;

use Behat\Mink\Element\NodeElement;
use Behat\Mink\Exception\ElementNotFoundException;
use Oro\Bundle\TestFrameworkBundle\Behat\Element\Element;
use WebDriver\Exception\ElementNotVisible;

class Grid extends Element
{
    /**
     * @param string $title
     * @throws \Exception
     */
    public function clickMassActionLink($title)
    {
        $massActionsButton = $this->getMassActionButton();
        $massActionsButton->press();
        $massActionLink = $this->getMassActionLink($title);
        $massActionLink->click();
    }

    /**
     * Get Element tr by row number
     *
     * @param int $rowNumber Number of grid record starting from 1
     * @return GridRow tr element of grid
     */
    public function getRowByNumber($rowNumber)
    {
        $rowIndex = $rowNumber - 1;
        $rows = $this->getRows();

        self::assertArrayHasKey(
            $rowIndex,
            $rows,
            sprintf('Can\'t get %s row, because there are only %s rows in grid', $rowNumber, count($rows))
        );

        return $rows[$rowIndex];
    }

    /**
     * Get Element tr by row content
     *
     * @param string $content Any content that can identify row
     * @return GridRow tr element of grid
     */
    public function getRowByContent($content)
    {
        $gridRow = $this->findElementContains('GridRow', $content);
        self::assertTrue($gridRow->isIsset(), sprintf('Grid has no record with "%s" content', $content));

        return $gridRow;
    }

    /**
     * @param int $number
     */
    public function checkFirstRecords($number)
    {
        $rows = $this->getRows();

        self::assertGreaterThan(
            $number,
            count($rows),
            sprintf('Can\'t check %s records, because grid has only %s records', $number, count($rows))
        );

        for ($i = 0; $i < $number; $i++) {
            $rows[$i]->checkMassActionCheckbox();
        }
    }

    /**
     * @param string $content
     */
    public function checkRecord($content)
    {
        $this->getRowByContent($content)->checkMassActionCheckbox();
    }

    /**
     * @return NodeElement
     * @throws \Exception
     */
    public function getMassActionButton()
    {
        $massActionsButton = $this->findButton('Mass Actions');

        if (!$massActionsButton || !$massActionsButton->isVisible()) {
            throw ElementNotVisible::factory(
                ElementNotVisible::ELEMENT_NOT_VISIBLE,
                'Mass Action dropdown is not prsent or not visible on page'
            );
        }

        return $massActionsButton;
    }

    /**
     * @param string $title
     * @return NodeElement|null
     */
    public function getMassActionLink($title)
    {
        return $this->elementFactory->createElement('GridFloatingMenu')->findLink($title);
    }

    /**
     * @param string $title
     * @throws \Exception
     */
    public function massCheck($title)
    {
        $this->elementFactory->createElement('MassActionHeadCheckbox')->click();
        $this->elementFactory->createElement('GridFloatingMenu')->clickLink($title);
    }

    /**
     * @param int $number
     * @throws ElementNotFoundException
     */
    public function selectPageSize($number)
    {
        $pageSizeElement = $this->elementFactory->createElement('PageSize');
        $pageSizeElement->find('css', '.btn')->click();
        $pageSizeElement->clickLink($number);
    }

    public function assertNoRecords()
    {
        self::assertCount(0, $this->getRows());
    }

    /**
     * @param string $content
     * @param string $action
     */
    public function clickActionLink($content, $action)
    {
        $row = $this->getRowByContent($content);
        $link = $this->getActionLink($action, $row);
        $link->click();
    }

    /**
     * @param $action
     * @param NodeElement $row
     * @return NodeElement
     * @throws ElementNotFoundException
     */
    public function getActionLink($action, NodeElement $row)
    {
        if ($showMoreLink = $row->find('named', ['link', '...'])) {
            $showMoreLink->mouseOver();
            $link = $this->elementFactory
                ->createElement('GridFloatingMenu')
                ->find('named', ['link', ucfirst($action)]);
        } else {
            $link = $row->find('named', ['link', $action]);
        }

        if (!$link) {
            throw new ElementNotFoundException($this->getDriver(), 'link', 'id|title|alt|text', $action);
        }

        return $link;
    }

    /**
     * @return GridRow[]
     */
    public function getRows()
    {
        return array_map(function (NodeElement $element) {
            return $this->elementFactory->wrapElement('GridRow', $element);
        }, $this->findAll('css', 'tbody tr'));
    }
}
